(*---------------------------------------------------------------------------*)
(* Basic library functions                                                   *)
(*---------------------------------------------------------------------------*)

(* structure Lib = struct *)

fun id x = x;
fun kcomb x y = x;

fun not b = if b then false else true;

fun fst p = 
 case p
  of (x,_) => x;

fun snd p = 
 case p
  of (_,y) => y;

fun the opt = case opt of SOME x => x;

fun funpow f n x = 
  if n <= 0 then x else funpow f (n-1) (f x);

fun inv_image rel f x y = rel (f x) (f y);

fun null l = (l=[]);

fun dest_cons list = 
 case list
  of h::t => (h,t)
   | _ => raise Bind;

fun hd list = 
 case list
  of h::_ => h
   | _ => raise Bind;

fun tl list = 
 case list
  of _::t => t
   | _ => raise Bind;

fun el n list = if n = 0 then hd list else el (n-1) (tl list);

fun mem x list = 
  case list
   of [] => false
    | h::t => (x=h) orelse mem x t;

fun append l1 l2 = 
 case l1
  of [] => l2
   | h::t => h::append t l2;

fun itlist f l z = 
 case l 
  of [] => z
   | h::t => f h (itlist f t z);

fun rev_itlist f l a = 
 case l 
  of [] => a
   | h::t => rev_itlist f t (f a h);

fun map f l = 
 case l 
  of [] => []
   | h::t => f h :: map f t;

fun rev l a = 
  case l
   of [] => a
    | h::t => rev t (h::a)

fun reverse l = rev l [];

fun len l a = 
  case l 
   of [] => a
    | _::t => len t (a+1);

fun length l = len l 0;

fun filter p l = 
 case l 
  of [] => []
   | h::t => if p h
               then h::filter p t
                else filter p t;

fun every p l = 
 case l 
  of [] => true
   | h::t => p h andalso every p t;

fun exists p l = 
 case l 
  of [] => false
   | h::t => p h orelse exists p t;

fun upto b t = 
 if b > t then [] else
 if b = t then [b] else
 b::upto (b+1) t;

fun genlist f n = if n<0 then [] else map f (upto 0 (n-1));

fun last l = 
 case l 
  of [] => raise Bind
   | [x] => x
   | h::t => last t;

fun front_last list = 
 case list
  of [] => raise Bind
   | [x] => ([],x)
   | h::t => case front_last t of (front,last) => (h::front,last);

fun end_itlist f list = 
 case front_last list  of (front,last) => itlist f front last;

fun drop n l = if n<=0 then l else drop (n-1) (tl l);

(*---------------------------------------------------------------------------*)
(* merge sort                                                                *)
(*---------------------------------------------------------------------------*)

fun merge rel l1 l2 = 
 case (l1,l2)
  of ([],[]) => []
   | ([],_) => l2
   | (_,[]) => l1
   | (x::t1,y::t2) => 
       if rel x y then 
          x::merge rel t1 l2 else 
          y::merge rel l1 t2;

fun sort2 rel x y = if rel x y then [x, y] else [y, x];

fun sort3 rel x y z =
 if rel x y then
     if rel y z then [x, y, z]
     else if rel x z then [x, z, y]
     else [z, x, y]
 else if rel y z then if rel x z then [y, x, z] else [y, z, x]
      else [z, y, x];

fun mergesortN rel n l = 
 if n = 0 then [] else
 if n = 1 then (if null l then [] else [hd l]) else
 if n = 2 
   then (case l 
          of (x::y::t) => sort2 rel x y
           | other => other) else 
 if n = 3 
   then (case l 
          of (x::y::z::t) => sort3 rel x y z
           | [x,y] => sort2 rel x y
           | other => other)
 else 
 let val len1 = n div 2
 in merge rel (mergesortN rel len1 l) 
              (mergesortN rel (n - len1) (drop len1 l))
 end;

fun mergesort rel l = mergesortN rel (length l) l;

(*---------------------------------------------------------------------------*)
(* Comparisons                                                               *)
(*---------------------------------------------------------------------------*)

datatype order = Equal | Less | Greater;

fun bool_cmp a b = if a=b then Equal else if a then Greater else Less;

fun int_cmp i j = if i=j then Equal else if i<j then Less else Greater;

fun list_cmp cmp l1 l2 = 
 case (l1,l2)
  of ([],[]) => Equal
   | ([],_) => Less
   | (_,[]) => Greater
   | (h1::t1,h2::t2) => 
      (case cmp h1 h2
        of Equal => list_cmp cmp t1 t2
         | verdict => verdict);

fun len_cmp l1 l2 = 
 if null l1
   then (if null l2 then Equal else Less)
   else (if null l2 then Greater else len_cmp(tl l1) (tl l2));

fun list_compare cmp l1 l2 = 
  case len_cmp l1 l2
   of Equal => list_cmp cmp l1 l2
    | Less => Less
    | Greater => Greater;

fun vector_cmp v1 v2 = 
 let fun chk i = 
     case (Vector.sub v1 i, Vector.sub v2 i)
      of (false,true) => Less
       | (true,false) => Greater
       | (true,true) => chk (i+1)
       | (false,false) => chk (i+1)
 in
  if v1=v2 then Equal 
  else
  case int_cmp (Vector.length v1) (Vector.length v2)
  of Less => Less
   | Greater => Greater
   | Equal => chk 0
end;

(* end; structure Lib *)

(*---------------------------------------------------------------------------*)
(* Finite maps, from examples/balanced_mapTheory.                            *)
(*---------------------------------------------------------------------------*)

structure Finite_Map = 
struct

datatype ('k,'v) balanced_map 
  = Tip 
  | Bin of int * 'k * 'v * ('k,'v)balanced_map * ('k,'v) balanced_map;

val ratio = 2;
val delta  = 3;

val empty = Tip;

fun singleton k x = Bin(1,k,x,Tip,Tip);

fun size fmap = 
 case fmap 
  of Tip => 0
   | Bin (s,k,v,l,r) => s;

fun member cmp k fmap = 
 case fmap
  of Tip => false
   | Bin (s,k', v, l, r) =>
     case cmp k k' 
      of Less => member cmp k l
       | Greater => member cmp k r
       | Equal => true;

fun lookup cmp k fmap = 
 case fmap 
  of Tip => NONE
   | Bin(s, k', v, l, r) =>
      case cmp k k' 
       of Less => lookup cmp k l
        | Greater => lookup cmp k r
        | Equal => SOME v;

fun foldrWithKey f z fmap = 
 case fmap
  of Tip => z
   | Bin(u, kx, x, l, r) => 
        foldrWithKey f (f kx x (foldrWithKey f z r)) l;

fun balanceL k x fmap1 fmap2 = 
 case (fmap1,fmap2) 
  of (Tip,Tip) 
      => Bin (1, k, x, Tip, Tip)
   | (Bin (s', k', v', Tip, Tip), Tip) 
      => Bin (2, k, x, Bin (s', k', v', Tip, Tip), Tip)
   | (Bin (u1, lk, lx, Tip, Bin(u2, lrk, lrx, u3, u4)), Tip) 
      => Bin(3, lrk, lrx, Bin(1, lk, lx, Tip, Tip), Bin(1,k, x, Tip, Tip))
   | (Bin(u5, lk, lx, Bin (s', k', v', l', r'), Tip), Tip) 
      => Bin(3, lk, lx, Bin(s', k', v', l', r'), Bin(1, k, x, Tip, Tip))
   | (Bin(ls, lk, lx, Bin(lls, k', v', l', r'), Bin(lrs, lrk, lrx, lrl, lrr)), Tip) 
      => if lrs < ratio*lls 
          then
            Bin (1+ls, lk, lx, Bin(lls, k', v', l', r'),
                 Bin (1+lrs, k, x, Bin (lrs, lrk, lrx, lrl, lrr), Tip))
          else
            Bin (1+ls, lrk, lrx, 
                 Bin (1+lls+size lrl, lk, lx, Bin(lls, k', v', l', r'), lrl),
                 Bin (1+size lrr, k, x, lrr, Tip))
   | (Tip, Bin(rs, k', v', l', r')) 
      => Bin (1+rs, k, x, Tip, Bin(rs, k', v', l', r'))
   | (Bin(ls, lk, lx, ll, lr), Bin(rs, k', v', l', r')) 
      => if ls > delta*rs  
          then (case (ll, lr) 
                 of (Bin (lls, u6, u7, u8, u9), Bin(lrs, lrk, lrx, lrl, lrr))
                    => if lrs < ratio*lls then
                        Bin (1+ls+rs, lk, lx, ll, 
                             Bin (1+rs+lrs, k, x, lr, Bin(rs, k', v', l', r')))
                      else
                        Bin (1+ls+rs, lrk, lrx, 
                             Bin (1+lls+size lrl, lk, lx, ll, lrl),
                             Bin (1+rs+size lrr, k, x, lrr, Bin (rs, k', v', l', r')))
                  | (u10, u11) => Tip (* error "Failure in Data.Map.balanceL" *)
               )
         else
            Bin (1+ls+rs, k, x, Bin (ls, lk, lx, ll, lr), Bin(rs, k', v', l', r'));

fun balanceR k x fmap1 fmap2 = 
 case (fmap1,fmap2) 
  of (Tip,Tip) 
      => Bin(1, k, x, Tip, Tip)
   | (Tip, Bin(s', k', v', Tip, Tip)) 
     => Bin(2, k, x, Tip, Bin(s', k', v', Tip, Tip))
   | (Tip,Bin(u1, rk, rx, Tip, Bin(s', k', v', l', r'))) 
     => Bin(3, rk, rx, Bin(1, k, x, Tip, Tip), Bin(s', k', v', l', r'))
   | (Tip, Bin(u2, rk, rx, Bin(u3, rlk, rlx, u4, u5), Tip))
      => Bin(3, rlk, rlx, 
             Bin(1, k, x, Tip, Tip), 
             Bin(1, rk, rx, Tip, Tip))
   | (Tip, Bin(rs, rk, rx, Bin(rls, rlk, rlx, rll, rlr), 
                           Bin(rrs, k', v', l', r')))
     => if rls < ratio*rrs 
        then
          Bin (1+rs, rk, rx,
               Bin (1+rls,k, x, Tip, Bin(rls, rlk, rlx, rll, rlr)),
               Bin(rrs, k', v', l', r'))
        else
          Bin (1+rs, rlk, rlx, 
               Bin (1+size rll, k, x, Tip, rll),
               Bin (1+rrs+size rlr, rk, rx, rlr, Bin(rrs, k', v', l', r')))
   | (Bin(ls, k', v', l', r'), Tip) 
     =>  Bin (1+ls, k, x, Bin(ls, k', v', l', r'), Tip)
   | (Bin(ls, k', v', l', r'), Bin(rs, rk, rx, rl, rr)) 
     => if rs > delta*ls 
        then
         (case (rl, rr) 
           of (Bin(rls, rlk, rlx, rll, rlr), Bin(rrs, u6, u7,u8,u9)) 
               => if rls < ratio*rrs 
                  then Bin (1+ls+rs, rk, rx,
                            Bin (1+ls+rls, k, x, Bin(ls, k', v', l', r'), rl),
                            rr)
                  else Bin (1+ls+rs, rlk, rlx,
                            Bin (1+ls+size rll, k, x, Bin(ls, k', v', l', r'), rll),
                            Bin (1+rrs+size rlr, rk, rx, rlr, rr))
            | (u10,u11) => Tip (* error "Failure in Data.Map.balanceR" *)
          )      
        else
           Bin (1+ls+rs, k, x, 
                Bin(ls, k', v', l', r'), Bin(rs, rk,rx,rl, rr));

fun insert cmp k v fmap = 
 case fmap 
  of Tip => singleton k v
   | Bin(s, k', v', l, r) =>
       case cmp k k' 
        of Less => balanceL k' v' (insert cmp k v l) r
         | Greater => balanceR k' v' l (insert cmp k v r)
         | Equal => Bin(s,k,v,l,r);

end; (* structure Finite_Map *)

(*---------------------------------------------------------------------------*)
(* Let's get started!                                                        *)
(*---------------------------------------------------------------------------*)

val alphabet_size = 128;
val alphabet = genlist Char.chr alphabet_size;

(*---------------------------------------------------------------------------*)
(* Character sets                                                            *)
(*---------------------------------------------------------------------------*)

val charset_empty = Vector.fromList (genlist (kcomb false) alphabet_size);
val charset_full = Vector.fromList (genlist (kcomb true) alphabet_size);

fun charset_mem c cs = Vector.sub cs (Char.ord c);

fun charset_sing c = 
 let val index = Char.ord c
     fun indexFn i = (i=index)
 in 
   Vector.fromList (genlist indexFn alphabet_size)
 end;

fun charset_union cs1 cs2 = 
 if cs1 = cs2 then cs1 else 
 if cs1 = charset_empty then cs2 else
 if cs2 = charset_empty then cs1 else
 let fun indexFn i = Vector.sub cs1 i orelse Vector.sub cs2 i
 in Vector.fromList (genlist indexFn alphabet_size)
 end;

fun charset_insert ch cset = 
  if charset_mem ch cset 
    then cset
  else charset_union (charset_sing ch) cset;

datatype regexp 
    = Chset of bool Vector.vector
    | Cat of regexp * regexp
    | Star of regexp
    | Or of regexp list
    | Neg of regexp;

(*---------------------------------------------------------------------------*)
(* Charset < Cat < Star < Or < Neg                                           *)
(*---------------------------------------------------------------------------*)

fun regexp_compare r1 r2 = 
 case (r1,r2)
  of (Chset cs1, Chset cs2) => vector_cmp cs1 cs2
   | (Chset cs, r) => Less
   | (Cat (r,s), Chset _) => Greater
   | (Cat(r1,r2),Cat(r3,r4)) =>
         (case regexp_compare r1 r3
           of Less => Less
            | Greater => Greater
            | Equal => regexp_compare r2 r4)
   | (Cat(r,s), _) => Less
   | (Star r, Chset _) => Greater
   | (Star r, Cat (s,t)) => Greater
   | (Star r,Star s) => regexp_compare r s
   | (Star r, _) => Less
   | (Or rs,Chset _) => Greater
   | (Or rs,Cat (r,s)) => Greater
   | (Or rs,Star _) => Greater
   | (Or rs1,Or rs2) => list_compare regexp_compare rs1 rs2
   | (Or rs, _) => Less
   | (Neg r,Neg s) =>  regexp_compare r s
   | (Neg r,_) => Greater;

fun regexp_leq r1 r2 =  
  case regexp_compare r1 r2
   of Less => true
    | Equal => true
    | Greater => false;

fun build_char_set cs = Chset cs;

val empty = Chset charset_empty;
val sigma = Chset charset_full;
val epsilon = Star (Chset charset_empty);

(*---------------------------------------------------------------------------*)
(* Support for smart derivatives                                             *)
(*---------------------------------------------------------------------------*)

fun assoc_cat r s = 
  case r 
   of Cat(r1,r2) => Cat(r1,assoc_cat r2 s)
    | otherwise => Cat(r,s);

fun build_cat r1 r2 =
   if (r1 = empty) orelse (r2 = empty) then empty else 
   if r1 = epsilon then r2 else 
   if r2 = epsilon then r1
   else assoc_cat r1 r2;

fun build_neg r = 
  case r
   of Neg r' => r'
    | otherwise => Neg r;

fun build_star r =
  case r
   of Star r' => r
    | otherwise => Star r;

fun flatten_or rlist = 
 case rlist
  of [] => []
   | Or rs::rs' => append rs (flatten_or rs')
   | r::rs => r :: flatten_or rs;

(*---------------------------------------------------------------------------*)
(* Requires all charsets to be leftmost in rs. Mergesort takes care of that. *)
(*---------------------------------------------------------------------------*)

fun merge_charsets list = 
 case list 
  of (Chset a::Chset b::t) => merge_charsets (Chset(charset_union a b)::t)
   | otherwise => list;

fun remove_dups rlist = 
 case rlist
  of [] => []
   | [r] => [r]
   | r1::r2::rs =>
      if regexp_compare r1 r2 = Equal then
        remove_dups (r2::rs)
      else
        r1::remove_dups (r2::rs);

fun build_or rlist =
 let val rlist' = 
         remove_dups 
           (merge_charsets 
              (mergesort regexp_leq 
                 (flatten_or rlist)))
 in 
  if mem (Neg empty) rlist' then Neg empty  
  else
  case rlist'
   of [] => empty
    | [r] => r
    | [Chset cs, r] => if cs = charset_empty then r else Or rlist'
    | Chset cs :: t => if cs = charset_empty then Or t else Or rlist'
    | _ => Or rlist'
end;

fun nullable r = 
 case r 
  of Chset cs => false
   | Star r'    => true
   | Cat (r,s) => nullable r andalso nullable s
   | Or rs     => exists nullable rs
   | Neg r'    => not(nullable r');

(*---------------------------------------------------------------------------*)
(* Take derivative, then apply smart constructors.                           *)
(*---------------------------------------------------------------------------*)

fun smart_deriv c r = 
 case r 
  of Chset cs => if charset_mem c cs then epsilon else empty
   | Cat (r,s) =>
      let val dr = build_cat (smart_deriv c r) s
      in if nullable r 
          then build_or [dr, smart_deriv c s]
          else dr
      end
   | Star r => build_cat (smart_deriv c r) (build_star r)
   | Or rs  => build_or (map (smart_deriv c) rs)
   | Neg r  => build_neg (smart_deriv c r);

fun normalize r = 
 case r 
  of Chset cs => build_char_set cs
   | Cat (s,t) => build_cat (normalize s) (normalize t)
   | Star r' => build_star (normalize r')
   | Or rs   => build_or (map normalize rs)
   | Neg r'  => build_neg (normalize r');

fun transitions r = map (fn c => (c,smart_deriv c r)) alphabet;

fun extend_states next_state state_map trans translist = 
 case translist
  of [] => (next_state,state_map,trans)
   | (c,r')::t =>
       case Finite_Map.lookup regexp_compare r' state_map 
        of SOME n => extend_states next_state state_map ((c,n)::trans) t
         | NONE   => extend_states (next_state + 1) 
                        (Finite_Map.insert regexp_compare 
                                   r' next_state state_map)
                        ((c,next_state)::trans)
                        t;

fun build_table translist r arg = 
 case arg of (next_state,state_map,table) =>
 case extend_states 
       next_state state_map [] translist of (next_state,state_map,trans) =>
 case Finite_Map.lookup regexp_compare r state_map 
  of SOME n => (next_state, state_map, (n,trans)::table)
   | NONE   => (next_state + 1,
                Finite_Map.insert regexp_compare r next_state state_map,
                (next_state,trans)::table);
 
fun insert_regexp r seen = Finite_Map.insert regexp_compare r () seen;

fun mem_regexp r seen = Finite_Map.member regexp_compare r seen;

(*---------------------------------------------------------------------------*)
(* Core regexp compiler. The seen argument holds the already-seen regexps,   *)
(* which become states in the final DFA. The n argument is a step-index used *)
(* to ensure that the function terminates.                                   *)
(*---------------------------------------------------------------------------*)

fun brzozo seen worklist acc n = 
 case worklist 
  of [] => let val p = (seen,acc) in SOME p end
   | r::t =>
     if n <= 0 then NONE else
     if mem_regexp r seen 
       then brzozo seen t acc (n-1)
        else let val translist = transitions r
             in brzozo (insert_regexp r seen) 
                    (append (remove_dups (map snd translist)) t)
                    (build_table translist r acc)
                    (n-1)
             end;

val bigIndex = 2147483647;

fun get_accepts fmap = 
 mergesort (fn x => fn y => x <= y)
  (Finite_Map.foldrWithKey 
      (fn r => fn n => fn nlist => if nullable r then n::nlist else nlist)
      [] fmap);

fun alist_to_vec list default n max =
 if n=0 then [] else
 case list
  of [] => default :: alist_to_vec [] default (n-1) max
   | ((x,y)::t) => 
      if n <= max then
        (if x = max - n then y :: alist_to_vec t default (n-1) max
         else if x < max - n then alist_to_vec t default n max
         else default :: alist_to_vec list default (n-1) max)
      else [];

fun accepts_to_finals_vector accept_states max_state = 
  alist_to_vec (map (fn x => (x,true)) accept_states) 
               false max_state max_state;

fun leq_on_keys p1 p2 = inv_image (fn x => fn y => x <= y) fst p1 p2;

fun table_to_vectors table =
 map (fn p => case p of (k,table2) => 
       alist_to_vec 
           (mergesort leq_on_keys 
              (map (fn p1 => case p1 of (c,n) => (Char.ord c,SOME n)) table2))
           NONE (alphabet_size-1) (alphabet_size-1))
     (mergesort leq_on_keys table);

fun compile_regexp r = 
 let val r' = normalize r 
 in case the (brzozo Finite_Map.empty [r'] 
                     (1, Finite_Map.singleton r' 0, []) bigIndex) 
    of (states,(last_state,state_numbering,table)) =>
        let val delta_vecs = table_to_vectors table
            val finals = accepts_to_finals_vector 
                            (get_accepts state_numbering) last_state
        in (state_numbering, delta_vecs, finals)
        end
 end;

fun exec_dfa final_states table n string = 
 case string
  of [] => el n final_states
   | c::t => 
       case el (Char.ord c) (el n table)
        of NONE => false
         | SOME k => exec_dfa final_states table k t;

(*---------------------------------------------------------------------------*)
(* Returns a function of type :string -> bool                                *)
(*---------------------------------------------------------------------------*)

fun matcher r = 
 case compile_regexp r  
  of (state_numbering,delta,finals) =>
     let val start_state = the (Finite_Map.lookup regexp_compare 
                                      (normalize r) state_numbering)
     in fn s => exec_dfa finals delta start_state (String.explode s)
     end;

(*---------------------------------------------------------------------------*)
(*    Tests                                                                  *)
(*---------------------------------------------------------------------------*)


fun charset_of s = itlist charset_insert (String.explode s) charset_empty;

val dot = Chset charset_full;
val digit = charset_of "0123456789";
val alpha = charset_of "abcdefghijklmnopqrstuvwxyz_ABCDEFGHIJKLMNOPQRSTUVWXYZ";
val alphanum = charset_union digit alpha;

fun cat r s = Cat(r,s);
fun listcat rlist = end_itlist cat rlist;
fun regexp_and r s = Neg(Or[Neg r, Neg s]);

fun stringlit s = 
 case map (fn c => Chset (charset_sing c)) (String.explode s)
  of [] => epsilon
   | clist => listcat clist;

fun regexp_pow r n = 
 if n=0 then epsilon else
 if n=1 then r 
 else Cat(r,regexp_pow r (n-1));

fun span r i j = 
 if j<i then [] else
 if i=j then [regexp_pow r i]
 else regexp_pow r i :: span r (i+1) j;

fun regexp_range r i j = Or (span r i j);

val test = matcher (stringlit"foobar");
test "fo2b";
test "foobar";
test "foobar1";

val test = matcher (Star (Chset digit));
test"";
test"1";
test"11434123412341234235456337467456745675256245";
test "a";
test "_[";

val test = matcher (regexp_range (Chset alphanum) 1 20);
test "a";
test "";
test "foobar";

(* `.*1` *)
val test = matcher (Cat(Star dot,stringlit"1"));
test"asdfasdfasd1";
test"";
test"1";

(* `[0-9]` *)
val test = matcher (Chset digit);
test "";
test "1";
test "9";
test "0";
test "10";

(*  `[0-9]*` *)
val test = matcher (Star(Chset digit));
test "";
test "1";
test "9";
test "0";
test "10";
test " a";
test "1024563735355365673463";

(* `(.*1)(12)*` *)
val test = matcher (Cat(Cat(Star dot,stringlit"1"), Star(stringlit"12")));
test "adfasd11212";
test"";

(* `b*|b*(a|ab*a)b*` *)
val a = stringlit"a";
val b = stringlit"b";
val ab = listcat [a,b];

val pat = Or[Star b, listcat [Star b, Or[a,listcat[a,Star b,a]], Star b]];
val test = matcher pat;
test "";
test "bbbb";
test "bbbbabb";
test "apha";
test "a";
test "baa";

(* `b*ab*ab*` *)
val pat = listcat [Star b,a,Star b,a, Star b];
val test = matcher pat;
test"bbbaa";
test"aa";
test"bababb";

(* `[]*|.|..|...` *)
val pat = Or [epsilon, dot,listcat[dot,dot], listcat[dot,dot,dot]];
val test = matcher pat;
test"";
test"a";
test"abb";
test"123";
test"1234";

(* `.|(ab)*|(ba)*` *)
val pat = Or [dot,Star(listcat[a,b]), Star (listcat[b,a])];
val test = matcher pat;
test"";
test"a";
test"7";
test"abba";
test"abb";
test"ababababab";
test"babababab";
test"bababababa";


(* `~((.*aa.* )|(.*bb.* ))` *)
val pat = Neg(Or[listcat[Star dot,a,a,Star dot],listcat[Star dot,b,b,Star dot]]);
val test = matcher pat;
             (true  = test ("")) 
   andalso   (true  = test ("a"))
   andalso   (true  = test ("b"))
   andalso   (false = test ("aa")) 
   andalso   (true  = test ("ab")) 
   andalso   (true  = test ("ba")) 
   andalso   (false = test ("bb")) 
   andalso   (true  = test ("ababababababababababababababababababababababababababab"))
   andalso   (false = test ("abababababababababababbababababababababababababababab"));


(* `(.*00.* )&~(.*01)` *)
val pat = regexp_and (listcat[Star dot,stringlit"00",Star dot])
                     (Neg (listcat[Star dot,stringlit"01"]));
val test = matcher pat;
              (true  = test ("00"))
    andalso  (false = test ("001"))
    andalso  (true  = test ("0111010101010111111000000"))
    andalso  (true  = test ("011101010101011111100000010101000111111111111111111111"))
    andalso  (true  = test ("0011010101010111111000000101010001111111111111111111110"))
    andalso  (false = test ("0011010101010111111000000101010001111111111111111111101"))
   ;

(*---------------------------------------------------------------------------*)
(* All strings with at least three consecutive ones and not ending in 01 or  *)
(*   consisting of all ones.                                                 *)
(* `(.*111.* )&~((.*01)|1* )`;                                               *)
(*---------------------------------------------------------------------------*)

val pat = regexp_and (listcat[Star dot,stringlit"111",Star dot])
                     (Neg (Or[listcat[Star dot,stringlit"01"],Star(stringlit"1")]));
val test = matcher pat;

             (true = test ("01110"))
    andalso  (false = test ("1"))
    andalso (false = test ("11"))
    andalso (false = test ("111"))
    andalso (false = test ("1111111111111111111111111111111111"))
    andalso (false = test ("11111111111111111111111111111111111111111111111111111111"))
    andalso (false = test ("1111111111111111111111111111111111111111111111111111111111111111"))
    andalso (true = test ("0111010101010111111000000"))
    andalso (true = test ("01101010101011111100000010101000111111111111111111111"))
    andalso (true = test ("10001101010101011000000101010001111111111111111111110"))
    andalso (false = test ("0011010101010111111000000101010001111111111111111111101"))
   ;

(*---------------------------------------------------------------------------*)
(* Hard cases                                                                *)
(*---------------------------------------------------------------------------*)

matcher (regexp_range (Chset alphanum) 1 20);
matcher (regexp_range (Chset alphanum) 1 50);
(*
matcher (regexp_range (Chset alphanum) 1 70);
matcher (regexp_range (Chset alphanum) 1 100);
*)
